//    DiskMod is a disk filter driver that allow creaion of page file.
//    Copyright (C) 2009  Panot Joonkhiaw
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <ntddk.h>
//#include <ntddstor.h>
//#include <ntdddisk.h>
#include <ntstrsafe.h>
//#include <wdmguid.h>

#if DBG
 #define DBGPrint0(xx) DbgPrint xx
 #define DBGPrint1(xx) DbgPrint xx
 #define DBGPrint2(xx) DbgPrint xx
 #define DBGPrint3(xx) 0
 #define DBGPrint4(xx) 0
 #define DBGCODE(xx) xx
#else
 #define DBGPrint0(xx) DbgPrint xx
 #define DBGPrint1(xx) 0
 #define DBGPrint2(xx) 0
 #define DBGPrint3(xx) 0
 #define DBGPrint4(xx) 0
 #define DBGCODE(xx) 0
#endif

enum {
    DiskModPoolTag = 'doMD',
};

typedef struct _DISKMOD_DEV_EXT {
	PDEVICE_OBJECT devobj;
	PDEVICE_OBJECT lowerPDO;
	LONG  pagefileusage;
	LONG  lowerpagefileusage;
} DISKMOD_DEV_EXT, *PDISKMOD_DEV_EXT;


// function declarations
NTSTATUS DriverEntry( IN PDRIVER_OBJECT  DriverObject, IN PUNICODE_STRING  RegistryPath );
DRIVER_UNLOAD     DiskModUnload; 
DRIVER_ADD_DEVICE DiskModAddDevice;

DRIVER_DISPATCH   DiskModDispatchSkip;
DRIVER_DISPATCH   DiskModDispatchPnP;
DRIVER_DISPATCH   DiskModDispatchPower;
DRIVER_DISPATCH   DiskModDispatchDeviceControl;
DRIVER_DISPATCH   DiskModDispatchSystemControl;
DRIVER_DISPATCH   DiskModDispatchCreateClose;
DRIVER_DISPATCH   DiskModDispatchReadWrite;
DRIVER_DISPATCH   DiskModDispatchSCSI;

IO_COMPLETION_ROUTINE DiskModIoCompletionPnPDeviceUsageNotification;

// ----------------------------------------
// non-pagable functions

NTSTATUS DiskModDispatchSkip( IN PDEVICE_OBJECT  DeviceObject, IN PIRP  Irp )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	IoSkipCurrentIrpStackLocation(Irp);
	return IoCallDriver(fdoext->lowerPDO, Irp);
}

NTSTATUS DiskModDispatchPower( IN PDEVICE_OBJECT  DeviceObject, IN PIRP  Irp )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	PIO_STACK_LOCATION iost;
	NTSTATUS status;
	iost = IoGetCurrentIrpStackLocation(Irp);
	switch (iost->MinorFunction)
	{
	case IRP_MN_QUERY_POWER:	//Optional, top-down
	case IRP_MN_SET_POWER:
		PoStartNextPowerIrp(Irp); // must call this for each IRP_MN_QUERY_POWER, IRP_MN_SET_POWER in Win2K-2003
		// then pass IRP down
		break;
	}
	// default -- pass IRP down the I/O stack
	IoSkipCurrentIrpStackLocation(Irp);
	status = PoCallDriver(fdoext->lowerPDO, Irp);
	return status;
}

// ----------------------------------------
// pagable functions

NTSTATUS DiskModIoCompletionPnPDeviceUsageNotification( IN PDEVICE_OBJECT  DeviceObject, IN PIRP  Irp, IN PVOID  Context )
{
	if (Irp->PendingReturned) 
		return STATUS_MORE_PROCESSING_REQUIRED;
	else {
		PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
		// override lower driver status to allow paging
		if (NT_SUCCESS(Irp->IoStatus.Status)) {
			// remember that lower device complete with success status
			InterlockedIncrement(&fdoext->lowerpagefileusage);
		} else {
			// override status
			Irp->IoStatus.Status = STATUS_SUCCESS;
		}
		return STATUS_SUCCESS;
	}
}

NTSTATUS DiskModDispatchPnP( IN PDEVICE_OBJECT  DeviceObject, IN PIRP  Irp )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	PIO_STACK_LOCATION iost;
	NTSTATUS status;
	KdPrint(("DiskMod: DispatchPnP\n"));
	iost = IoGetCurrentIrpStackLocation(Irp);
	status = Irp->IoStatus.Status;
	switch (iost->MinorFunction)
	{
	case IRP_MN_START_DEVICE:			//Required, bottom-up
		KdPrint(("DiskMod:  START_DEVICE\n"));
		break;
		//IoCopyCurrentIrpStackLocationToNext(Irp);
		//IoSetCompletionRoutine(Irp, DiskModIoCompletionPnPStartDevice, NULL, TRUE,TRUE,TRUE);
		//return IoCallDriver(fdoext->lowerPDO, Irp);
	case IRP_MN_QUERY_STOP_DEVICE:		//Required, top-down
		KdPrint(("DiskMod:  QUERY_STOP_DEVICE\n"));
		break;
		// Win98: Stop for remove, Win2K+: Stop for rebalancing resource. 
	case IRP_MN_STOP_DEVICE:			//Required, top-down
		KdPrint(("DiskMod:  STOP_DEVICE\n"));
		break;
	case IRP_MN_CANCEL_STOP_DEVICE:		//Required, bottom-up
		KdPrint(("DiskMod:  CANCEL_STOP_DEVICE\n"));
		break;
		//IoCopyCurrentIrpStackLocationToNext(Irp);
		//IoSetCompletionRoutine(Irp, DiskModIoCompletionPnPCancelStop, NULL, TRUE,TRUE,TRUE);
		//return IoCallDriver(fdoext->lowerPDO, Irp);
	case IRP_MN_QUERY_REMOVE_DEVICE:	//Required, top-down
		KdPrint(("DiskMod:  QUERY_REMOVE_DEVICE\n"));
		break;
	case IRP_MN_SURPRISE_REMOVAL:		//Required, top-down
		KdPrint(("DiskMod:  SURPRISE_REMOVAL\n"));
		break;
	case IRP_MN_REMOVE_DEVICE:			//Required, top-down, synchronize
		KdPrint(("DiskMod:  REMOVE_DEVICE\n"));
		// call lower driver
		IoSkipCurrentIrpStackLocation(Irp);
		status = IoCallDriver(fdoext->lowerPDO, Irp);
		// delete device
		IoDetachDevice(fdoext->lowerPDO);
		IoDeleteDevice(DeviceObject);
		KdPrint(("DiskMod:  REMOVE_DEVICE end\n"));
		return status;
	case IRP_MN_CANCEL_REMOVE_DEVICE:	//Required, bottom-up
		KdPrint(("DiskMod:  CANCEL_REMOVE_DEVICE\n"));
		break;
		//IoCopyCurrentIrpStackLocationToNext(Irp);
		//IoSetCompletionRoutine(Irp, DiskModIoCompletionPnPCancelRemove, NULL, TRUE,TRUE,TRUE);
		//return IoCallDriver(fdoext->lowerPDO, Irp);
	//case IRP_MN_QUERY_CAPABILITIES:			//Optional
	//	break;
	//case IRP_MN_QUERY_PNP_DEVICE_STATE:		//Optional
	//	break;
	case IRP_MN_FILTER_RESOURCE_REQUIREMENTS:	//Optional
	//	KdPrint(("DiskMod:  FILTER_RESOURCE_REQUIREMENTS\n"));
		break;
	case IRP_MN_DEVICE_USAGE_NOTIFICATION:	//Required
		KdPrint(("DiskMod:  DEVICE_USAGE_NOTIFICATION\n"));
		Irp->IoStatus.Status = STATUS_SUCCESS;
		if (iost->Parameters.UsageNotification.Type == DeviceUsageTypePaging)
		{
			if (iost->Parameters.UsageNotification.InPath)
			{	// in path
				InterlockedIncrement(&fdoext->pagefileusage);
				IoCopyCurrentIrpStackLocationToNext(Irp);
				IoSetCompletionRoutine(Irp, DiskModIoCompletionPnPDeviceUsageNotification, NULL, TRUE,TRUE,TRUE);
				return IoCallDriver(fdoext->lowerPDO, Irp);
			} else {
				// not in path
				LONG count = InterlockedDecrement(&fdoext->pagefileusage);
				if (count >= fdoext->pagefileusage)
				{
					IoCompleteRequest(Irp, IO_NO_INCREMENT);
					return STATUS_SUCCESS;
				} else {
					InterlockedDecrement(&fdoext->lowerpagefileusage);
					// notify lower driver
					// Pass IRP down to lower driver
					break;
				}
			}
		}
		// non-paging file notification, just pass IRP down device stack
		break;
	case IRP_MN_QUERY_DEVICE_RELATIONS:		//Required, PASSIVE_LEVEL
		KdPrint(("DiskMod:  QUERY_DEVICE_RELATIONS\n"));
		//{
		//	switch (iost->Parameters.QueryDeviceRelations.Type)
		//	{
			//case BusRelations:			//Required
			//	KdPrint(("DiskMod:   BusRelation\n"));
			//	break;
			//case RemovalRelations:		//Optional
			//	break;
			//case EjectionRelations:		//No
			//case TargetDeviceRelation:	//No
			//	;
		//	}
		//}
		break;
	//case IRP_MN_QUERY_RESOURCES:				//No
	//case IRP_MN_QUERY_RESOURCE_REQUIREMENTS:	//No
	//case IRP_MN_QUERY_ID:						//No
	//case IRP_MN_QUERY_ID:
	//	KdPrint(("DiskMod:  QUERY_ID  IdType=%d FDO\n",iost->Parameters.QueryId.IdType));
	//	break;	
	//case IRP_MN_QUERY_DEVICE_TEXT:			//No
	//	KdPrint(("DiskMod:  QUERY_DEVICE_TEXT FDO\n"));
	//	break;
	//case IRP_MN_QUERY_BUS_INFORMATION:		//No
	//case IRP_MN_QUERY_INTERFACE:			//Optional
	//	KdPrint(("DiskMod:  QUERY_INTERFACE\n"));
	//	break;
	//case IRP_MN_READ_CONFIG:					//No
	//case IRP_MN_WRITE_CONFIG:					//No
	//case IRP_MN_EJECT:						//No
	//case IRP_MN_SET_LOCK:						//No
	//	break;
	}
	// default -- pass IRP down the I/O stack
	IoSkipCurrentIrpStackLocation(Irp);
	status = IoCallDriver(fdoext->lowerPDO, Irp);
	return status;
}

NTSTATUS DiskModAddDevice( IN PDRIVER_OBJECT  DriverObject, IN PDEVICE_OBJECT  PhysicalDeviceObject )
{
	PDEVICE_OBJECT fdo;
	NTSTATUS status;
	KdPrint(("DiskMod: AddDevice start\n"));
	status = IoCreateDevice( DriverObject, sizeof(DISKMOD_DEV_EXT), NULL /*no need for device name*/,
		FILE_DEVICE_BUS_EXTENDER, FILE_DEVICE_SECURE_OPEN, FALSE, &fdo);
	if (!NT_SUCCESS(status)) {
		KdPrint(("DiskMod: AddDevice create device failed status=0x%X\n",status));
		return status;
	} else {
		PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)(fdo->DeviceExtension);
		fdoext->devobj = fdo;
		fdoext->pagefileusage = 0;
		fdoext->lowerpagefileusage = 0;
		fdoext->lowerPDO = IoAttachDeviceToDeviceStack(fdo, PhysicalDeviceObject);
		if (!NT_SUCCESS(status)) {
			KdPrint(("DiskMod: AddDevice attach to device stack failed status=0x%X\n",status));
			IoDeleteDevice(fdo);
			return status;
		}
		// initialize complete
		fdo->Flags &= ~DO_DEVICE_INITIALIZING;
		status = STATUS_SUCCESS;
		KdPrint(("DiskMod: AddDevice success\n"));
	}
	return status;
}

VOID DiskModUnload( IN PDRIVER_OBJECT  DriverObject )
{
	// free driver-wide allocated resource
	KdPrint(("DiskMod: Unload ----------\n\n"));
}

// ----------------------------------------
// initialization functions

NTSTATUS DriverEntry( IN PDRIVER_OBJECT  DriverObject, IN PUNICODE_STRING  RegistryPath )
{
	KdPrint(("DiskMod: DriverEntry start ----------\n"));

	// initialize DriverObject
	DriverObject->DriverExtension->AddDevice = DiskModAddDevice;
	//DriverObject->DriverStartIo = DiskModStartIo; 
	DriverObject->DriverUnload  = DiskModUnload; 
	DriverObject->MajorFunction[IRP_MJ_CREATE] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_CLOSE ] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_READ  ] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_WRITE ] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_PNP   ] = DiskModDispatchPnP; 
	DriverObject->MajorFunction[IRP_MJ_POWER ] = DiskModDispatchPower; 
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_INTERNAL_DEVICE_CONTROL] = DiskModDispatchSkip;

	KdPrint(("DiskMod: DriverEntry success\n"));
	return STATUS_SUCCESS; 
}
